import codecs
import copy
import logging
import os
import tempfile
import threading
from Queue import Queue
from functools import wraps

from pyvertica.connection import get_connection


logger = logging.getLogger(__name__)


def require_started_batch(func):
    """
    Decorator to assure that the batch has been started before calling ``func``

    This decorator should only be used on class methods. As well this assumes
    that the class has an attribute called ``_in_batch`` (``bool``) containing
    the state and a method called ``_start_batch`` (without arguments). The
    latter will be called to start the batch.

    :return:
        Decorated ``func``.

    """
    @wraps(func)
    def inner_func(self, *args, **kwargs):
        if not self._in_batch:
            self._start_batch()
        return func(self, *args, **kwargs)
    return inner_func


class QueryThread(threading.Thread):
    """
    Thread object which will execute the ``COPY`` query.

    :param cursor:
        A :py:mod:`!pyodbc` cursor object.

    :param sql_query_str:
        A ``str`` representing the ``COPY`` query.

    :param semaphore_obj:
        An instance of :py:class:`!threading.Semaphore`.

    :param fifo_path:
        A ``str`` representing the path of the fifo file.

    :param exc_queue:
        An instance of class:`Queue.Queue` instance to put exceptions in.

    """

    daemon = True
    """
    If the script is interrupted (^C) it will kill all threads marked as daemon
    if only those are left.
    """

    def __init__(
            self, cursor, sql_query_str, semaphore_obj, fifo_path, exc_queue):
        super(QueryThread, self).__init__()
        self.cursor = cursor
        self.sql_query_str = sql_query_str
        self.semaphore_obj = semaphore_obj
        self.exc_queue = exc_queue
        self.fifo_path = fifo_path

    def run(self):
        """
        Handle executing the SQL query.

        This method will be called when starting this thread.

        """
        logger.debug('Thread started with SQL statement: {0}'.format(
            self.sql_query_str))
        try:
            self.cursor.execute(self.sql_query_str)
        except Exception as e:
            logger.exception('Something unexpected happened')

            # the exception will be re-raised in the main thread
            self.exc_queue.put(e)

            # we need to consume the fifo, to make sure it isn't blocking the
            # write (and thus hanging forever).
            for line in codecs.open(self.fifo_path, 'r', 'utf-8'):
                pass

        logger.debug('Thread done')
        self.semaphore_obj.release()


class VerticaBatch(object):
    """
    Object for writing multiple records to Vertica in a batch.

    Usage example::

        from pyvertica.batch import VerticaBatch

        batch = VerticaBatch(
            dsn='VerticaDWH',
            table_name='schema.my_table',
            truncate=True,
            column_list=['column_1', 'column_2'],
            copy_options={
                'DELIMITER': ',',
            }
        )

        row_list = [
            ['row_1_val_1', 'row_1_val_2'],
            ['row_2_val_1', 'row_2_val_2'],
            ...
        ]

        for column_data_list in row_list:
            batch.insert_list(column_data_list)

        error_bool, error_file_obj = batch.get_errors()

        if error_bool:
            print error_file_obj.read()

        batch.commit()

    .. note:: It is also possible to call :py:meth:`~.VerticaBatch.commit`
        multiple times (for example after every 50000 records). Please note
        that after the first insert and after calling
        :py:meth:`~.VerticaBatch.commit`, the output of
        :py:meth:`~.VerticaBatch.get_errors` will reflect the new serie of
        inserts and thus not contain the "old" inserts.

    .. note:: Creating a new batch object will not create a lock on the target
        table. This will happen only after first insert.

    .. note:: Although the batch object is automagically reusable, after a
        :py:meth:`~.VerticaBatch.commit` the locks are realeased up to next
        insert.


    :param table_name:
        A ``str`` representing the table name (including the schema) to write
        to. Example: ``'staging.my_table'``.

    :param odbc_kwargs:
        A ``dict`` containing the ODBC connection keyword arguments. E.g.::

            {
                'dsn': 'TestDSN',
            }

        .. seealso:: https://code.google.com/p/pyodbc/wiki/Module

    :param truncate_table:
        A ``bool`` indicating if the table needs truncating before first
        insert. Default: ``False``. *Optional*.

    :param reconnect:
        A ``bool`` passed to the connection object to decide if pyvertica
        should directly reconnect to a random node to bypass a load balancer.

    :param analyze_constraints:
        A ``bool`` indicating if a ``ANALYZE_CONSTRAINTS`` startement should
        be executed when getting errors. Default: ``True``. *Optional*.

    :param column_list:
        A ``list`` containing the columns that will be written. *Optional*.

    :param copy_options:
        A ``dict`` containing the keys to override. For a list of existing keys
        and their defaults, see :py:attr:`~.VerticaBatch.copy_options_dict`.
        *Optional*.

    """
    copy_options_dict = {
        'DELIMITER': ';',
        'ENCLOSED BY': '"',
        'SKIP': 0,
        'NULL': '',
        'RECORD TERMINATOR': '\x01',
        'NO COMMIT': True,
        'REJECTEDFILE': True,
    }
    """
    Default copy options for SQL query.
    """

    def __init__(
            self,
            table_name,
            odbc_kwargs={},
            truncate_table=False,
            reconnect=True,
            analyze_constraints=True,
            column_list=[],
            copy_options={}):
        # make sure we are not logging any passwords :)
        odbc_kwargs_copy = copy.deepcopy(odbc_kwargs)
        if 'password' in odbc_kwargs_copy:
            odbc_kwargs_copy['password'] = '*****'
        logger.debug(
            'Initializing VerticaBatch with odbc_kwargs={0}, table_name={1}, '
            'column_list={2}'.format(
                odbc_kwargs_copy, table_name, column_list))

        self._odbc_kwargs = odbc_kwargs
        self._table_name = table_name
        self._column_list = column_list
        self._analyze_constraints = analyze_constraints
        self.copy_options_dict.update(copy_options)

        self._total_count = 0
        self._batch_count = 0

        self._in_batch = False

        # setup db connection
        self._connection = get_connection(
            reconnect=reconnect, **self._odbc_kwargs)
        self._cursor = self._connection.cursor()

        # truncate table, if needed
        if truncate_table:
            self._truncate_table()

    def _truncate_table(self):
        """
        Truncate table.
        """
        logger.info('Truncating table {0}'.format(self._table_name))
        self._cursor.execute('TRUNCATE TABLE {0}'.format(self._table_name))

    def _start_batch(self):
        """
        Start the batch.

        This will create the FIFO file, a temporary file for the rejected
        inserts and this will setup and start the :py:class:`.QueryThread`.

        """
        self._in_batch = True
        self._batch_count = 0
        self._query_exc_queue = Queue()

        # create FIFO
        self._fifo_path = os.path.join(tempfile.mkdtemp(), 'fifo')
        os.mkfifo(self._fifo_path)

        # create rejected file obj
        self._rejected_file_obj = tempfile.NamedTemporaryFile(bufsize=0)

        # setup query thread
        self._query_thread_semaphore_obj = threading.Semaphore(0)
        self._query_thread = QueryThread(
            self._cursor,
            self._get_sql_lcopy_str(),
            self._query_thread_semaphore_obj,
            self._fifo_path,
            self._query_exc_queue,
        )
        self._query_thread.start()

        logger.debug('Opening FIFO')
        self._fifo_obj = codecs.open(self._fifo_path, 'w', 'utf-8')

        logger.debug('Batch started')

    @require_started_batch
    def _end_batch(self):
        """
        End the batch.

        This will remove the FIFO file and stop the :py:class:`.QueryThread`.

        """
        ended_clean = True

        logger.debug('Closing FIFO')
        self._fifo_obj.close()

        logger.debug('Waiting for thread to finish')
        self._query_thread_semaphore_obj.acquire()
        logger.debug('Thread finished')

        logger.debug('Terminating thread')
        self._query_thread.join(2)
        if self._query_thread.is_alive():
            ended_clean = False
            logging.error('Terminating thread timed out!')

        self._in_batch = False
        os.remove(self._fifo_path)
        os.rmdir(os.path.dirname(self._fifo_path))

        logger.debug('Batch ended')

        if not self._query_exc_queue.empty():
            raise self._query_exc_queue.get()

        return ended_clean

    def get_batch_count(self):
        """
        Return number (``int``) of inserted items since last commit.

        .. warning:: When using :py:meth:`~.VerticaBatch.insert_raw` this
            value represents the number of raw ``str`` objects inserted, not
            the number of lines!

        :return:
            An ``int``.

        """
        return self._batch_count

    def get_total_count(self):
        """
        Return total number (``int``) of inserted items.

        .. warning:: When using :py:meth:`~.VerticaBatch.insert_raw` this
            value represents the number of raw ``str`` objects inserted, not
            the number of lines!

        :return:
            An ``int``.

        """
        return self._total_count

    @require_started_batch
    def _get_sql_lcopy_str(self):
        """
        Get ``str`` representing the COPY query.

        :return:
            A ``str`` representing the query.

        """
        # table name
        output_str = 'COPY {0}'.format(self._table_name)

        # columns, if available
        if self._column_list:
            output_str += ' ({0})'.format(', '.join(self._column_list))

        # fifo path
        output_str += " FROM LOCAL '{0}'".format(self._fifo_path)

        # rejected file
        if self.copy_options_dict['REJECTEDFILE']:
            output_str += " REJECTED DATA '{0}'".format(
                self._rejected_file_obj.name)

        # other arguments which map one-to-one
        for key in [
                'DELIMITER',
                'ENCLOSED BY',
                'SKIP',
                'NULL',
                'RECORD TERMINATOR']:
            value = self.copy_options_dict[key]

            if isinstance(value, int):
                output_str += ' {0} {1}'.format(key, value)
            elif isinstance(value, str):
                output_str += " {0} '{1}'".format(key, value)

        # NO COMMIT statement, which needs to be at the end
        if self.copy_options_dict['NO COMMIT']:
            output_str += ' NO COMMIT'

        return output_str

    def insert_list(self, value_list):
        """
        Insert a ``list`` of values (instead of a ``str`` representing a line).

        Example::

            batch.insert_list(['value_1', 'value_2'])

        :param value_list:
            A ``list``. Each item should represent a column value.

        """
        enclosed_by = self.copy_options_dict['ENCLOSED BY']
        escaped_enclosed_by = '\\%s' % enclosed_by

        str_value_list = [
            '%s%s%s' % (
                enclosed_by,
                unicode(value).replace(enclosed_by, escaped_enclosed_by),
                enclosed_by
            )
            if value is not None
            else '' for value in value_list]

        insert_str = self.copy_options_dict['DELIMITER'].join(str_value_list)

        return self.insert_line(insert_str)

    @require_started_batch
    def insert_line(self, line_str):
        """
        Insert a ``str`` containing all the values.

        This is useful when inserting lines directly from a CSV file for
        example.

        .. note:: When you have a loghandler with ``DEBUG`` level, every query
            will be logged. For performance reason, this log statement is only
            executed when ``__debug__`` equals ``True`` (which is the default
            case). For a better performance, you should invoke the Python
            interpreter with the ``-O`` argument or set the environment
            variable ``PYTHONOPTIMIZE`` to something.

        Example::

            batch.insert_line('"value_1";"value_2"')

        :param line_str:
            A ``str`` representing the line to insert. Make sure the ``str``
            is formatted according :py:attr:`~.VerticaBatch.copy_options_dict`.
            Example: ``'"value1";"value2";"value3"'``.

        """
        if __debug__:
            logger.debug(u'Inserting line: {0}'.format(line_str))

        self._fifo_obj.write(
            line_str + self.copy_options_dict['RECORD TERMINATOR'])

        self._total_count += 1
        self._batch_count += 1

    @require_started_batch
    def insert_raw(self, raw_str):
        """
        Insert a raw ``str``.

        A raw ``str`` does not have to be a complete row, but can be a part of
        a row or even multiple rows. This is useful when you have a file that
        is already in a format readable by Vertica.
        """
        if __debug__:
            logger.debug(u'Inserting raw: {0}'.format(raw_str))

        self._fifo_obj.write(raw_str)

        self._total_count += 1
        self._batch_count += 1

    def get_errors(self):
        """
        Get errors that were raised since the last commit.

        This will check constraint errors and rejected data by the database.
        Please note that this will remove the rejected data file after calling
        this method. Therfore it is not possible to call this method more than
        once per batch!

        .. note:: Since this is checking the contraints as well, it is assumed
            that all contrains were met before starting the batch. Otherwise,
            these errors will show up within this method.

        :return:
            A ``tuple`` with as first item a ``bool`` representing if there are
            errors (``True`` = errors, ``False`` = no errors). The second item
            is a file-like object containing the error-data in plain text.
            Since this is an instance of :py:class:`!tempfile.TemporaryFile`,
            it will be removed automatically.

        """
        if self._in_batch:
            self._end_batch()

        error_file_obj = tempfile.TemporaryFile(bufsize=0)

        if not self.get_batch_count():
            return(False, error_file_obj)

        if self._analyze_constraints:
            try:
                analyze_constraints = self._cursor.execute(
                    "SELECT ANALYZE_CONSTRAINTS('{0}')".format(
                        self._table_name))
            except Exception as e:
                if not 'no constraints defined' in str(e).lower():
                    raise e
                analyze_constraints = None

            if analyze_constraints and analyze_constraints.rowcount > 0:
                error_file_obj.write(
                    'At least one constraint not met: {0}\n'.format(
                        ', '.join(analyze_constraints.fetchone())))

        self._rejected_file_obj.seek(0)
        file_size = os.path.getsize(self._rejected_file_obj.name)
        read_func = lambda: self._rejected_file_obj.read(1024 * 1024)
        error_prefix = 'Rejected data at line: '

        for counter, line in enumerate(iter((read_func), '')):
            if counter == 0:
                error_file_obj.write(error_prefix)

            line = line.replace(
                self.copy_options_dict['RECORD TERMINATOR'],
                '\n{0}'.format(error_prefix)
            )

            if self._rejected_file_obj.tell() == file_size:
                line = line[:-len(error_prefix)]

            error_file_obj.write(line)

        errors = error_file_obj.tell() > 0
        error_file_obj.seek(0)

        return (errors, error_file_obj)

    def commit(self):
        """
        Commit the current transaction.
        """
        batch_count = self.get_batch_count()

        if self._in_batch:
            self._end_batch()

        self._connection.commit()
        logger.info('Transaction comitted, {0} lines inserted'.format(
            batch_count))

    def get_cursor(self):
        """
        Return a cursor to the database.

        This is useful when you want to add extra data within the same
        transaction of the batch import.

        :return:
            Instance of :py:class:`!pyodbc.Cursor`.

        """
        return self._connection.cursor()



# -*- coding: utf-8 -*-

import os
import stat
import tempfile
import threading
import unittest2 as unittest
from Queue import Queue

from mock import Mock, patch

from pyvertica.batch import (
    QueryThread, VerticaBatch, require_started_batch)


class RequireStartedBatchDecoratorTestCase(unittest.TestCase):
    """
    Tests for :py:func:`.require_started_batch`.
    """
    def setUp(self):
        class TestClass(object):
            _start_batch = Mock()

            def __init__(self, in_batch):
                self._in_batch = in_batch

            @require_started_batch
            def test(self, foo):
                return foo

        self.TestClass = TestClass

    def test_not_in_batch(self):
        """
        Test when ``_in_batch`` is ``False``.
        """
        test_class = self.TestClass(False)
        self.assertEqual('foo', test_class.test('foo'))
        test_class._start_batch.assert_called_once_with()

    def test_in_batch(self):
        """
        Test when ``_in_batch`` is ``True``.
        """
        test_class = self.TestClass(True)
        self.assertEqual('foo', test_class.test('foo'))
        self.assertEqual(0, test_class._start_batch.call_count)


class QueryThreadTestCase(unittest.TestCase):
    """
    Tests for :py:class:`.QueryThread`.
    """
    def test___init__(self):
        """
        Test initialization of :py:class:`.QueryThread`.
        """
        cursor = Mock()
        sql_query_str = Mock()
        semaphore_obj = Mock()
        fifo_path = Mock()
        exc_queue = Mock()

        query_thread = QueryThread(
            cursor, sql_query_str, semaphore_obj, fifo_path, exc_queue)

        self.assertEqual(cursor, query_thread.cursor)
        self.assertEqual(sql_query_str, query_thread.sql_query_str)
        self.assertEqual(semaphore_obj, query_thread.semaphore_obj)
        self.assertEqual(fifo_path, query_thread.fifo_path)
        self.assertEqual(exc_queue, query_thread.exc_queue)

    def test_run(self):
        """
        Test :py:meth:`.QueryThread.run`.
        """
        cursor = Mock()
        sql_query_str = Mock()
        semaphore_obj = Mock()

        query_thread = QueryThread(
            cursor, sql_query_str, semaphore_obj, Mock(), Mock())
        query_thread.run()

        cursor.execute.assert_called_once_with(sql_query_str)
        semaphore_obj.release.assert_called_once_with()

    def test_run_raising_exception(self):
        """
        Test :py:meth:`.QueryThread.run` raising an exception.
        """
        file_obj = tempfile.NamedTemporaryFile(bufsize=0, delete=False)
        file_obj.write('foo\nbar\n')
        file_obj.close()
        semaphore_obj = threading.Semaphore(0)

        cursor = Mock()
        cursor.execute.side_effect = Exception('Boom!')
        exc_queue = Queue()

        QueryThread(
            cursor, Mock(), semaphore_obj, file_obj.name, exc_queue).start()
        semaphore_obj.acquire()

        os.remove(file_obj.name)
        self.assertTrue(isinstance(exc_queue.get(), Exception))


class VerticaBatchTestCase(unittest.TestCase):
    """
    Test for :py:class:`.VerticaBatch`.
    """
    def get_batch(self, **kwargs):
        arguments = {
            'odbc_kwargs': {'dsn': 'TestDSN'},
            'table_name': 'schema.test_table',
            'truncate_table': False,
            'column_list': [
                'column_1',
                'column_2',
                'column_3',
            ],
            'copy_options': {'DELIMITER': ',', 'SKIP': 1},
        }
        arguments.update(kwargs)
        return VerticaBatch(**arguments)

    @patch('pyvertica.batch.VerticaBatch._truncate_table')
    @patch('pyvertica.batch.get_connection')
    def test___init__no_truncate(self, get_connection, truncate_table):
        """
        Test initialization of :py:class:`.VerticaBatch` without truncate.
        """
        batch = self.get_batch()

        # variables
        self.assertEqual({'dsn': 'TestDSN'}, batch._odbc_kwargs)
        self.assertEqual('schema.test_table', batch._table_name)
        self.assertEqual(
            ['column_1', 'column_2', 'column_3'], batch._column_list)
        self.assertEqual({
            'DELIMITER': ',',
            'ENCLOSED BY': '"',
            'SKIP': 1,
            'RECORD TERMINATOR': '\x01',
            'NULL': '',
            'NO COMMIT': True,
            'REJECTEDFILE': True,
        }, batch.copy_options_dict)
        self.assertEqual(0, batch._total_count)
        self.assertEqual(0, batch._batch_count)
        self.assertFalse(batch._in_batch)

        # db connection
        get_connection.assert_called_once_with(dsn='TestDSN', reconnect=True)
        self.assertEqual(get_connection.return_value, batch._connection)
        batch._connection.cursor.assert_called_once_with()
        self.assertEqual(batch._connection.cursor.return_value, batch._cursor)

        # truncate
        self.assertEqual(0, truncate_table.call_count)

    @patch('pyvertica.batch.VerticaBatch._truncate_table')
    @patch('pyvertica.batch.get_connection')
    def test__init__truncate(self, get_connection, truncate_table):
        """
        Test truncate during initialization of :py:class:`.VerticaBatch`.
        """
        self.get_batch(truncate_table=True)
        truncate_table.assert_called_once_with()

    @patch('pyvertica.batch.get_connection')
    def test_truncate_table(self, get_connection):
        """
        Test :py:meth:`.VerticaBatch._truncate_table`.
        """
        batch = self.get_batch()
        batch._cursor = Mock()

        batch._truncate_table()
        batch._cursor.execute.assert_called_once_with(
            'TRUNCATE TABLE schema.test_table')

    @patch('pyvertica.batch.codecs')
    @patch('pyvertica.batch.VerticaBatch._get_sql_lcopy_str')
    @patch('pyvertica.batch.QueryThread')
    @patch('pyvertica.batch.threading')
    @patch('pyvertica.batch.get_connection')
    def test__start_batch(
                self,
                get_connection,
                threading,
                QueryThreadMock,
                get_sql_lcopy_str,
                codecs,
            ):
        """
        Test :py:meth:`.VerticaBatch._start_batch`.
        """
        batch = self.get_batch()
        batch._batch_count = 10

        batch._start_batch()

        self.assertTrue(batch._in_batch)
        self.assertEqual(0, batch._batch_count)

        # test files
        self.assertTrue(stat.S_ISFIFO(os.stat(batch._fifo_path).st_mode))
        self.assertTrue(os.path.exists(batch._rejected_file_obj.name))
        codecs.open.assert_called_once_with(batch._fifo_path, 'w', 'utf-8')
        self.assertEqual(codecs.open.return_value, batch._fifo_obj)

        # test thread setup
        threading.Semaphore.assert_called_once_with(0)
        self.assertEqual(
            threading.Semaphore.return_value,
            batch._query_thread_semaphore_obj
        )
        QueryThreadMock.assert_called_once_with(
            batch._cursor,
            batch._get_sql_lcopy_str.return_value,
            batch._query_thread_semaphore_obj,
            batch._fifo_path,
            batch._query_exc_queue,
        )
        self.assertEqual(
            QueryThreadMock.return_value,
            batch._query_thread
        )
        batch._query_thread.start.assert_called_once_with()

    @patch('pyvertica.batch.os.remove')
    @patch('pyvertica.batch.os.rmdir')
    @patch('pyvertica.batch.get_connection')
    @patch('pyvertica.batch.VerticaBatch._start_batch')
    def test__end_batch_clean(
            self, start_batch, get_connection, remove, rmdir):
        """
        Test :py:meth:`.VerticaBatch._end_batch` ending clean.

        :py:mod:`!time` is mocked to speedup the tests since there is no need
        to actually wait.

        """
        query_thread = Mock()
        query_thread.is_alive.return_value = False

        batch = self.get_batch()
        batch._fifo_path = '/tmp/abcd1234/fifo'
        batch._fifo_obj = Mock()
        batch._query_thread_semaphore_obj = Mock()
        batch._query_thread = query_thread
        batch._query_exc_queue = Mock()
        batch._query_exc_queue.empty.return_value = True

        end_return = batch._end_batch()

        batch._fifo_obj.close.assert_called_once_with()
        batch._query_thread_semaphore_obj.acquire.assert_called_once_with()

        query_thread.join.assert_called_once_with(2)
        query_thread.is_alive.assert_called_once_with()

        os.remove('/tmp/abcd1234/fifo')
        os.rmdir('/tmp/abcd1234')

        self.assertFalse(batch._in_batch)
        self.assertTrue(end_return)

    @patch('pyvertica.batch.os.remove')
    @patch('pyvertica.batch.os.rmdir')
    @patch('pyvertica.batch.VerticaBatch._start_batch')
    @patch('pyvertica.batch.get_connection')
    def test__end_batch_dirty(
            self, get_connection, start_batch, rmdir, remove):
        """
        Test :py:meth:`.VerticaBatch._end_batch` ending 'dirty'.
        """
        cursor = get_connection.return_value.cursor.return_value
        cursor.execute.return_value.fetchone.return_value = 1

        query_thread = Mock()
        query_thread.is_alive.return_value = True

        batch = self.get_batch()
        batch._fifo_path = '/tmp/abcd1234/fifo'
        batch._fifo_obj = Mock()
        batch._query_thread_semaphore_obj = Mock()
        batch._query_thread = query_thread
        batch._query_exc_queue = Mock()
        batch._query_exc_queue.empty.return_value = True

        os.remove('/tmp/abcd1234/fifo')
        os.rmdir('/tmp/abcd1234')

        self.assertFalse(batch._end_batch())
        self.assertFalse(batch._in_batch)

    @patch('pyvertica.batch.get_connection')
    def test_get_batch_count(self, get_connection):
        """
        Test :py:meth:`.VerticaBatch.get_batch_count`.
        """
        batch = self.get_batch()
        batch._batch_count = Mock()

        self.assertEqual(batch._batch_count, batch.get_batch_count())

    @patch('pyvertica.batch.get_connection')
    def test_get_total_count(self, get_connection):
        """
        Test :py:meth:`.VerticaBatch.get_total_count`.
        """
        batch = self.get_batch()
        batch._total_count = Mock()

        self.assertEqual(batch._total_count, batch.get_total_count())

    @patch('pyvertica.batch.VerticaBatch._start_batch')
    @patch('pyvertica.batch.get_connection')
    def test__get_sql_lcopy_str(self, get_connection, start_batch):
        """
        Test :py:meth:`.VerticaBatch._get_sql_lcopy_str`.
        """
        batch = self.get_batch()
        batch._fifo_path = '/tmp/fifo'
        batch._rejected_file_obj = Mock()
        batch._rejected_file_obj.name = '/tmp/rejected'

        self.assertEqual(
            "COPY schema.test_table (column_1, column_2, column_3) "
            "FROM LOCAL '/tmp/fifo' REJECTED DATA '/tmp/rejected' "
            "DELIMITER ',' ENCLOSED BY '\"' SKIP 1 NULL '' "
            "RECORD TERMINATOR '\x01' NO COMMIT",
            batch._get_sql_lcopy_str()
        )

    @patch('pyvertica.batch.get_connection')
    def test_insert_list(self, get_connection):
        """
        Test :py:meth:`.VerticaBatch.insert_list`.
        """
        batch = self.get_batch()
        batch.insert_line = Mock()

        self.assertEqual(
            batch.insert_line.return_value,
            batch.insert_list(
                [u'valué1', 'valu"e2', 'None', None, 100, 'value4'])
        )
        batch.insert_line.assert_called_once_with(
            u'"valué1","valu\\"e2","None",,"100","value4"')

    @patch('pyvertica.batch.VerticaBatch._start_batch')
    @patch('pyvertica.batch.get_connection')
    def test_insert_line(self, get_connection, start_batch):
        """
        Test :py:meth:`.VerticaBatch.insert_line`.
        """
        batch = self.get_batch()
        batch._fifo_obj = Mock()

        batch.insert_line('"value1";"value2";"value3"')

        batch._fifo_obj.write.assert_called_once_with(
            '"value1";"value2";"value3"\x01')
        self.assertEqual(1, batch._total_count)
        self.assertEqual(1, batch._batch_count)

    @patch('pyvertica.batch.VerticaBatch._start_batch')
    @patch('pyvertica.batch.get_connection')
    def test_insert_raw(self, get_connection, start_batch):
        """
        Test :py:meth:`.VerticaBatch.insert_raw`.
        """
        batch = self.get_batch()
        batch._fifo_obj = Mock()

        raw_mock = Mock()

        batch.insert_raw(raw_mock)
        batch._fifo_obj.write.assert_called_with(raw_mock)
        self.assertEqual(1, batch._total_count)
        self.assertEqual(1, batch._batch_count)

    @patch('pyvertica.batch.get_connection')
    def test_get_errors_exception(self, get_connection):
        """
        Test :py:meth:`.VerticaBatch.get_errors` raising exception.
        """
        batch = self.get_batch()
        batch.get_batch_count = Mock(return_value=10)
        batch._in_batch = True
        batch._end_batch = Mock()
        batch._cursor = Mock()

        batch._cursor.execute.side_effect = Exception('Kaboom!')

        self.assertRaises(Exception, batch.get_errors)
        batch._end_batch.assert_called_once_with()

    @patch('pyvertica.batch.get_connection')
    def test_get_errors_no_constraints(self, get_connection):
        """
        Test :py:meth:`.VerticaBatch.get_errors` without constraints.
        """
        batch = self.get_batch()
        batch.get_batch_count = Mock(return_value=10)
        batch._end_batch = Mock()
        batch._cursor = Mock()
        batch._rejected_file_obj = tempfile.NamedTemporaryFile()

        batch._cursor.execute.side_effect = Exception(
            'Somewhere are no constraints defined, oh dear!')

        self.assertFalse(batch.get_errors()[0])
        self.assertEqual(0, batch._end_batch.call_count)

    @patch('pyvertica.batch.get_connection')
    def test_get_errors_disabled_analyze_constraints(self, get_connection):
        """
        Test :py:meth:`.VerticaBatch.get_errors` when analyze_contraints=False.
        """
        batch = self.get_batch(analyze_constraints=False)
        batch.get_batch_count = Mock(return_value=10)
        batch._end_batch = Mock()
        batch._cursor = Mock()
        batch._rejected_file_obj = tempfile.NamedTemporaryFile()

        self.assertEqual(0, batch._cursor.execute.call_count)

    @patch('pyvertica.batch.get_connection')
    def test_get_errors_constraint_error(self, get_connection):
        """
        Test :py:meth:`.VerticaBatch.get_errors` with constraint errors.
        """
        batch = self.get_batch()
        batch.get_batch_count = Mock(return_value=10)
        batch._end_batch = Mock()
        batch._cursor = Mock()
        batch._rejected_file_obj = tempfile.NamedTemporaryFile()

        batch._cursor.execute.return_value.fetchone.return_value = ['a', 'b']

        errors_tuple = batch.get_errors()

        self.assertTrue(errors_tuple[0])
        self.assertEqual(
            'At least one constraint not met: a, b\n', errors_tuple[1].read())

    @patch('pyvertica.batch.get_connection')
    def test_get_errors_rejected(self, get_connection):
        """
        Test :py:meth:`.VerticaBatch.get_errors` with rejected errors.
        """
        batch = self.get_batch()
        batch.get_batch_count = Mock(return_value=10)
        batch._end_batch = Mock()
        batch._cursor = Mock()
        batch._rejected_file_obj = tempfile.NamedTemporaryFile()

        batch._cursor.execute.return_value.rowcount = 0

        with open(batch._rejected_file_obj.name, 'w') as f:
            f.write('123\x01456\x01')

        errors_tuple = batch.get_errors()

        self.assertTrue(errors_tuple[0])
        self.assertEqual(
            'Rejected data at line: 123\nRejected data at line: 456\n',
            errors_tuple[1].read()
        )

    @patch('pyvertica.batch.tempfile')
    @patch('pyvertica.batch.get_connection')
    def test_get_errors_no_batch_count(self, get_connection, tempfile):
        """
        Test :py:meth:`.VerticaBatch.get_errors` without batch count.
        """
        batch = self.get_batch()
        batch.get_batch_count = Mock(return_value=0)
        batch._end_batch = Mock()

        self.assertEqual(
            (False, tempfile.TemporaryFile.return_value),
            batch.get_errors()
        )

    @patch('pyvertica.batch.get_connection')
    def test_commit_in_batch(self, get_connection):
        """
        Test :py:meth:`.VerticaBatch.commit` while in batch.
        """
        batch = self.get_batch()
        batch._in_batch = True
        batch._end_batch = Mock()
        batch._connection = Mock()

        batch.commit()

        batch._end_batch.assert_called_once_with()
        batch._connection.commit.assert_called_with()

    @patch('pyvertica.batch.get_connection')
    def test_commit_not_in_batch(self, get_connection):
        """
        Test :py:meth:`.VerticaBatch.commit` while not in batch.
        """
        batch = self.get_batch()
        batch._end_batch = Mock()
        batch._connection = Mock()

        batch.commit()

        self.assertEqual(0, batch._end_batch.call_count)
        batch._connection.commit.assert_called_once_with()

    @patch('pyvertica.batch.get_connection')
    def test_get_cursor(self, get_connection):
        """
        Test :py:meth:`.VerticaBatch.get_cursor`.
        """
        batch = self.get_batch()
        batch._connection = Mock()

        self.assertEqual(
            batch._connection.cursor.return_value, batch.get_cursor())
        batch._connection.cursor.assert_called_once_with()



import pyodbc


def get_connection(reconnect=True, **kwargs):
    """
    Get :py:mod:`!pyodbc` connection for the given ``dsn``.

    Usage example::

        from pyvertica.connection import get_connection


        connection = get_connection('TestDSN')
        cursor = connection.cursor()

    The connection will be made in two steps (with the assumption that you are
    connection via a load-balancer). The first step is connecting to the
    load-balancer and selecting a random node address. Then it will connect
    to that specific node and return this connection instance. This is done
    to avoid that all the data has to pass the load-balancer.

    .. note:: Depending on the given keyword arguments, you need to have
        a ``odbc.ini`` file on your system.

    :param reconnect:
        A ``boolean`` asking to reconnect to skip load balancer.


    :param kwargs:
        Keyword arguments accepted by the :py:mod:`!pyodbc` module.
        See: http://code.google.com/p/pyodbc/wiki/Module#connect

    :return:
        Return an instance of :class:`!pyodbc.Connection`.

    """
    connection = pyodbc.connect(**kwargs)

    if reconnect:
        return get_connection(
            reconnect=False,
            servername=_get_random_node_address(connection),
            **kwargs
        )

    return connection


def _get_random_node_address(connection):
    """
    Return the address of a random node in the cluster.

    :param connection:
        An instance of :class:`!pyodbc.Connection`.

    :return:
        A ``str`` representing the address of the node.

    """
    cursor = connection.cursor()
    cursor.execute(
        'SELECT node_address FROM nodes WHERE node_state = ? '
        'ORDER BY RANDOM() LIMIT 1',
        'UP'
    )
    return cursor.fetchone().node_address


def connection_details(con):
    """
    Given one connection objects returns information about it.

    :param con:
        An instance of :class:`!pyodbc.Connection`.

    return:
        A ``dict`` with the following keys / values:

        host
            Connected node IP address

        user
            Connected username

        db
            Connected database name

    """
    details = con.execute('''
        SELECT
            n.node_address as host
          , CURRENT_USER() as user
          , CURRENT_DATABASE() as db
        FROM v_monitor.current_session cs
        JOIN v_catalog.nodes n ON n.node_name=cs.node_name
        ''').fetchone()
    return {
        'host': details.host,
        'user': details.user,
        'db': details.db
    }



import logging
from datetime import datetime

from pyvertica.connection import get_connection
from pyvertica.batch import VerticaBatch


logger = logging.getLogger(__name__)


class AlreadyImportedError(Exception):
    """
    Exception is raised when trying to import the same source twice.
    """
    pass


class BatchImportError(Exception):
    """
    Exception is raised when the batch import is returning errors.
    """
    pass


class BaseImporter(object):
    """
    Base class for importing data into Vertica.

    Note, before using this base importer, make sure you have created the
    history table for batch imports::

        CREATE TABLE meta.batch_history (
            batch_source_name VARCHAR(255),
            batch_source_type_name VARCHAR(255),
            batch_source_path VARCHAR(255),
            batch_import_timestamp TIMESTAMP
        )

    Usage example::

        class AdGroupPerformanceReportImporter(BaseImporter):
            table_name = 'adwords_ad_group_performance'
            batch_source_name = 'adwords_api'
            batch_source_type_name = 'ad_group_performance_report'

            mapping_list = (
                {
                    'field_name': 'AccountCurrencyCode',
                    'db_field_name': 'account_currency_code',
                    'db_data_type': 'VARCHAR(10)',
                },
                {
                    'field_name': 'AccountDescriptiveName',
                    'db_field_name': 'account_descriptive_name',
                    'db_data_type': 'VARCHAR(512)',
                },
                {
                    'field_name': 'AccountTimeZoneId',
                    'db_field_name': 'account_time_zone_id',
                    'db_data_type': 'VARCHAR(100)',
                },
                ...
            )

        iterable_object = [
            {
                'AccountCurrencyCode': 'EUR',
                'AccountDescriptiveName': 'Test account description',
                'AccountTimeZoneId': '(GMT+01:00) Amsterdam,'
            },
            ...
        ]

        report_importer = AdGroupPerformanceReportImporter(
            iterable_object,
            dsn='VerticaTST',
            schema_name='test',
            batch_source_path='ADGROUP_PERFORMANCE_REPORT.1234.20120521',
        )
        report_importer.start_import()

    In the example above, we are importing a ``list`` of dicts. More likely,
    this ``iterable_object`` would be a reader class for your data-source,
    which is iterable and would return a ``dict`` with the expected fields`.


    :param reader_obj:
        An object that is iterable and returns for every line the data as a
        ``dict``.

        .. note:: Passing a ``list`` of ``dict`` objects will work as well.

    :param odbc_kwargs:
        A ``dict`` containing the ODBC connection keyword arguments. E.g.::

            {
                'dsn': 'TestDSN',
            }

        .. seealso:: https://code.google.com/p/pyodbc/wiki/Module


    :param schema_name:
        Name of the DB schema to use.

    :param batch_source_path:
        A ``str`` describing the path to the source. This can be a file path
        when importing from a file, or an identifier when the source is an
        API. This should be unique for every import!

    :param kwargs:
        Optional extra keyword arguments, will be stored as ``self._kwargs``.

    """

    table_name = ''
    """
    Name of the database table (excluding the schema) (``str``).
    """

    batch_history_table = 'meta.batch_history'
    """
    Name of the database table containing the batch history (including the
    schema name) (``str``). The structure of this table is::

            batch_source_name VARCHAR(255)
            batch_source_type_name VARCHAR(255)
            batch_source_path VARCHAR(255)
            batch_import_timestamp TIMESTAMP

    """

    batch_source_name = ''
    """
    The name of the source which the data is retrieved from. E.g.: for AdWords,
    this this could be something like ``'adwords_api'``.
    """

    batch_source_type_name = ''
    """
    The type of data that is imported from the source. E.g.: for the AdWords
    API, this could be something like ``ADGROUP_PERFORMANCE_REPORT``.
    """

    mapping_list = ()
    """
    A ``tuple`` of ``dict`` objects to map record columns to db columns.

    The fields specified in this list might be a sub-set of the available
    fields in the record. Only specify the fields you want to store in the DB.

    Each ``dict`` must contain the following keys:

    ``field_name``
        A ``str`` representing the field name as it is in the ``dict``
        containing the data (as returned by the ``reader_obj``).

    ``db_data_type``
        A ``str`` representing the field type in the database,
        eg: ``'varchar(10)'``.

    Optionally, each ``dict`` can contain the following keys:

    ``db_field_name``
        A ``str`` representing the field name within the database. This only
        needs to be set when it does not match with the source field name.

    """

    extra_fields = (
        {
            'field_name': 'batch_source_name',
            'db_data_type': 'VARCHAR(255)',
        },
        {
            'field_name': 'batch_source_path',
            'db_data_type': 'VARCHAR(255)',
        },
        {
            'field_name': 'batch_import_timestamp',
            'db_data_type': 'TIMESTAMP',
        },
    )
    """
    A ``tuple`` of ``dict`` objects to prepend fields to the data.

    This list enables the possibility to add extra data to the database, for
    example data related to the import (source name, source identifier, import
    timestamp, ...).

    Each ``dict`` must contain the following keys:

    ``field_name``
        A ``str`` representing the DB field name.

    ``db_data_type``
        A ``str`` representing the field type in the database,
        eg: ``'varchar(10)'``.

    Then, for every field, you should define a method within your class
    which is is named following this template:
    ``get_extra_{field_name}_data``. This method will be called for every
    imported record with a ``dict`` containing the row data.

    .. warning:: Make sure there is no collision between these fields and the
        fields defined in :py:attr:`~.BaseImporter.mapping_list`.

    """

    _batch_import_timestamp = None

    def __init__(
            self,
            reader_obj,
            schema_name,
            batch_source_path,
            odbc_kwargs={},
            **kwargs):
        self._reader_obj = reader_obj
        self._odbc_kwargs = odbc_kwargs
        self._schema_name = schema_name
        self._kwargs = kwargs
        self._kwargs.update({
            'batch_source_path': batch_source_path,
        })
        logger.debug('{0} initialized'.format(self.__class__.__name__))

    def _get_vertica_batch(self):
        """
        Return a VerticaBatch instance.

        :return:
            Instance of :py:class:`.VerticaBatch`.

        """
        logger.info('Setup VerticaBatch for {0}'.format(
            self.__class__.__name__))
        return VerticaBatch(
            odbc_kwargs=self._odbc_kwargs,
            table_name='{0}.{1}'.format(self._schema_name, self.table_name),
            column_list=self._get_db_column_list(),
        )

    def _get_db_column_list(self):
        """
        Return a list of DB column names.

        First this will take the field names from
        :py:attr:`~.BaseImporter.extra_fields`. Then it will take the fields
        defined in :py:attr:`~.BaseImporter.mapping_list`.

        :return:
            A ``list`` of ``str`` objects.

        """
        extra_field_list = [x['field_name'] for x in self.extra_fields]
        data_field_list = [x.get('db_field_name', x['field_name'])
            for x in self.mapping_list]

        db_column_list = extra_field_list + data_field_list

        logger.info('DB columns for {0}: {1}'.format(
            self.__class__.__name__, db_column_list))
        return db_column_list

    def _get_row_value_list(self, row_data_dict):
        """
        Get list of row values which can be inserted into the DB.

        :param row_data_dict:
            A ``dict`` containing the fields and their values for one data-row.

        :return:
            A ``list`` of values, in the right column order (as generated
            by :py:meth:`~.BaseImporter._get_db_column_list`).

        """
        output_list = []
        for field_dict in self.extra_fields:
            data_method = getattr(self, 'get_extra_{0}_data'.format(
                field_dict['field_name']))
            output_list.append(data_method(row_data_dict))

        output_list.extend([row_data_dict[x['field_name']]
            for x in self.mapping_list])

        return output_list

    def _insert_into_history(self, db_cursor):
        """
        Insert import instance into batch history.

        This is used to determine the next time which objects are and aren't
        imported.

        :param db_cursor:
            An instance of :py:class:`!pyodbc.Cursor`.

        """
        db_cursor.execute(
            'INSERT INTO {batch_history_table} (batch_source_name, '
            'batch_source_type_name, batch_source_path, '
            'batch_import_timestamp) VALUES (?, ?, ?, ?)'.format(
                batch_history_table=self.batch_history_table
            ),
            self.batch_source_name,
            self.batch_source_type_name,
            self.get_extra_batch_source_path_data(None),
            self.get_extra_batch_import_timestamp_data(None),
        )

    def get_sql_create_table_statement(self):
        """
        Return SQL statement for creating the DB table.

        This will first use the fields specified in
        :py:attr:`~.BaseImporter.extra_fields`, followed by the fields in
        :py:attr:`~.BaseImporter.mapping_list`.

        :return:
            A ``str`` representing the SQL statement.

        """
        db_field_list = []

        for field_dict in self.extra_fields:
            db_field_list.append('{field_name} {db_data_type}'.format(
                **field_dict))

        for field_dict in self.mapping_list:
            db_field_name = field_dict.get(
                'db_field_name', field_dict['field_name'])
            db_field_list.append('{0} {1}'.format(
                db_field_name, field_dict['db_data_type']))

        return 'CREATE TABLE {schema}.{table_name} ({fields})'.format(
            schema=self._schema_name,
            table_name=self.table_name,
            fields=', '.join(db_field_list)
        )

    def start_import(self):
        """
        Start the import.

        This will import all the data from the ``reader_obj`` argument
        (given when constructing :py:class:`.BaseImporter`). In case there
        are no errors, it will commit the import at the end.

        :raises:
            :py:exc:`.BatchImportError` when there are errors during the
            import. Errors are logged to the logger object.

        :raises:
            :py:exc:`.AlreadyImportedError` when there already an import exists
            with the same batch-source path. Before starting your import, you
            can test this by
            calling :py:meth:`~.BaseImporter.get_batch_source_path_exists`.

        """
        batch_source_path_exists = self.get_batch_source_path_exists(
            self._kwargs['batch_source_path'], odbc_kwargs=self._odbc_kwargs)

        if batch_source_path_exists:
            raise AlreadyImportedError(
                'There is already an import with '
                'batch_source_path={0}'.format(
                    self._kwargs['batch_source_path']
                )
            )

        batch_obj = self._get_vertica_batch()

        for data_dict in self._reader_obj:
            batch_obj.insert_list(self._get_row_value_list(data_dict))

        logger.info('Last line inserted')

        errors_bool, errors_file_obj = batch_obj.get_errors()
        if not errors_bool:
            batch_db_cursor = batch_obj.get_cursor()
            self._insert_into_history(batch_db_cursor)
            batch_obj.commit()
        else:
            for error_line in errors_file_obj:
                logger.error('Batch error ({0}): {1}'.format(
                    self._kwargs['batch_source_path'],
                    error_line.rstrip('\r\n')))

            raise BatchImportError(
                'Errors detected during the import of '
                'batch_source_path={0}'.format(
                    self._kwargs['batch_source_path']
                )
            )

    @classmethod
    def get_batch_source_path_exists(cls, batch_source_path, odbc_kwargs={}):
        """
        Check if the batch source-path exists in the database.

        :param batch_source_path:
            The batch source-path (``str``).

        :param odbc_kwargs:
            A ``dict`` containing the ODBC connection keyword arguments. E.g.::

                {
                    'dsn': 'TestDSN',
                }

            .. seealso:: https://code.google.com/p/pyodbc/wiki/Module

        :return:
            ``True`` if it already exists, else ``False``.

        """
        connection = get_connection(**odbc_kwargs)
        cursor = connection.cursor()
        cursor.execute(
            'SELECT batch_source_path FROM {batch_history_table} '
            'WHERE batch_source_name = ? AND batch_source_type_name = ? AND '
            'batch_source_path = ? LIMIT 1'.format(
                batch_history_table=cls.batch_history_table
            ),
            cls.batch_source_name,
            cls.batch_source_type_name,
            batch_source_path
        )
        row = cursor.fetchone()

        if row:
            return True
        return False

    @classmethod
    def get_last_imported_batch_source_path(cls, odbc_kwargs):
        """
        Return the last imported batch source-path.

        :param odbc_kwargs:
            A ``dict`` containing the ODBC connection keyword arguments. E.g.::

                {
                    'dsn': 'TestDSN',
                }

            .. seealso:: https://code.google.com/p/pyodbc/wiki/Module

        :return:
            A ``str`` representing the last imported batch source-path.

        """
        connection = get_connection(**odbc_kwargs)
        cursor = connection.cursor()
        cursor.execute(
            'SELECT batch_source_path FROM {batch_history_table} '
            'WHERE batch_source_name = ? AND batch_source_type_name = ? '
            'ORDER BY batch_import_timestamp DESC LIMIT 1'.format(
                batch_history_table=cls.batch_history_table
            ),
            cls.batch_source_name,
            cls.batch_source_type_name
        )
        row = cursor.fetchone()

        if row:
            return row[0]
        return None

    def get_extra_batch_source_name_data(self, row_data_dict):
        """
        Return batch source name.

        :param row_data_dict:
            A ``dict`` containing the row-data.

        :return:
            The value set in :py:attr:`~.BaseImporter.batch_source_name`.

        """
        return self.batch_source_name

    def get_extra_batch_source_path_data(self, row_data_dict):
        """
        Return batch source path.

        :param row_data_dict:
            A ``dict`` containing the row-data.

        :return:
            A ``str`` containing the batch-source path (this is given as the
            ``batch_source_path`` argument on constructing this class).

        """
        return self._kwargs['batch_source_path']

    def get_extra_batch_import_timestamp_data(self, row_data_dict):
        """
        Return batch import timestamp.

        :param row_data_dict:
            A ``dict`` containing the row-data.

        :return:
            A ``str`` in ISO 8601 format, with a space a separator.

        """
        if not self._batch_import_timestamp:
            self._batch_import_timestamp = datetime.utcnow().isoformat(' ')
        return self._batch_import_timestamp



import unittest2 as unittest

from mock import Mock, call, patch

from pyvertica.connection import get_connection
from pyvertica.connection import _get_random_node_address


class ModuleTestCase(unittest.TestCase):
    """
    Tests for :py:mod:`~pyvertica.connection`.
    """
    @patch('pyvertica.connection._get_random_node_address')
    @patch('pyvertica.connection.pyodbc')
    def test_get_connection(self, pyodbc, get_random_node_address):
        """
        Test :py:func:`.get_connection`.
        """
        pyodbc.connect.side_effect = ['connection1', 'connection2']

        connection = get_connection(dsn='TestDSN', foo='bar', bar='foo')

        self.assertEqual([
            call(dsn='TestDSN', foo='bar', bar='foo'),
            call(
                dsn='TestDSN',
                servername=get_random_node_address.return_value,
                foo='bar',
                bar='foo',
            ),
        ], pyodbc.connect.call_args_list)

        get_random_node_address.assert_called_once_with('connection1')

        self.assertEqual('connection2', connection)

    def test__get_random_node_address(self):
        """
        Test :py:func:`._get_random_node_address`.
        """
        connection = Mock()
        cursor = connection.cursor()
        row = cursor.fetchone.return_value

        self.assertEqual(
            row.node_address,
            _get_random_node_address(connection)
        )

        cursor.execute.assert_called_once_with(
            'SELECT node_address FROM nodes WHERE node_state = ? '
            'ORDER BY RANDOM() LIMIT 1',
            'UP'
        )



import unittest2 as unittest

import pyodbc

# Sidable logging during unittests
import logging
logging.disable(logging.CRITICAL)

from subprocess import CalledProcessError
from mock import Mock, call, patch
from pyvertica.migrate import VerticaMigrator, VerticaMigratorError


class VerticaMigratorConnection(unittest.TestCase):
    """
    Test for the connection steps of :py:class:`.VerticaMigrator`.
    Another class will test the rest, mocking out globally the connections
    """


class VerticaMigratorTestCase(unittest.TestCase):
    """
    Test for :py:class:`.VerticaMigrator`.
    """

    @patch('pyvertica.migrate.VerticaMigrator._set_connections')
    @patch('pyvertica.migrate.VerticaMigrator._sanity_checks')
    def get_migrator(self, sanity_checks, set_connections, **kwargs):
        migrator = VerticaMigrator('SourceDSN', 'TargetDSN', False, **kwargs)
        return migrator

    ### different sanity options.
    @patch('pyvertica.migrate.VerticaMigrator._source_ip', '1.2.3.4',
        create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target_ip', '5.6.7.8',
        create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._set_connections')
    def test__sanity_ok(self, cnx, target):
        """
        Test :py:meth:`.VerticaMigrator._sanity_checks`.
        No exception should be raised if IP are different.
        """
        target.execute.return_value.fetchone.return_value = [0]
        VerticaMigrator('SourceDSN', 'TargetDSN', False)

    @patch('pyvertica.migrate.VerticaMigrator._source_ip', '1.2.3.4',
        create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target_ip', '1.2.3.4',
        create=True)
    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._set_connections')
    def test__sanity_same_ip_diff_db(self, cnx, target, source):
        """
        Test :py:meth:`.VerticaMigrator._sanity_checks`.
        No exception should be raised if IPs are identical but
        DBs are different.
        """
        target.execute.return_value.fetchone.side_effect = [['targetDB'], [0]]
        source.execute.return_value.fetchone.return_value = ['sourceDB']
        VerticaMigrator('SourceDSN', 'TargetDSN', False)

    @patch('pyvertica.migrate.VerticaMigrator._source_ip', '1.2.3.4',
        create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target_ip', '1.2.3.4',
        create=True)
    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._set_connections')
    def test__sanity_same_ip_same_db(self, cnx, target, source):
        """
        Test :py:meth:`.VerticaMigrator._sanity_checks`.
        Exception if IPs are identical and DBs are identical.
        """
        target.execute.return_value.fetchone.side_effect = [['targetDB'], [0]]
        source.execute.return_value.fetchone.return_value = ['targetDB']
        self.assertRaises(VerticaMigratorError,
        VerticaMigrator, 'SourceDSN', 'TargetDSN', False)

    @patch('pyvertica.migrate.VerticaMigrator._source_ip', '1.2.3.4',
        create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target_ip', '5.6.7.8',
        create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._set_connections')
    def test__sanity_not_empty(self, cnx, target):
        """
        Test :py:meth:`.VerticaMigrator._sanity_checks`.
        Exception if target DB is not empty.
        """
        target.execute.return_value.fetchone.return_value = [42]
        self.assertRaises(VerticaMigratorError,
            VerticaMigrator, 'SourceDSN', 'TargetDSN', False)

    @patch('pyvertica.migrate.VerticaMigrator._source_ip', '1.2.3.4',
        create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target_ip', '5.6.7.8',
        create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._set_connections')
    def test__sanity_not_empty_but_thats_ok(self, cnx, target):
        """
        Test :py:meth:`.VerticaMigrator._sanity_checks`.
        No exception if target DB is not empty and we know it:
         even_not_empty=True.
        """
        target.execute.return_value.fetchone.return_value = [42]
        VerticaMigrator('SourceDSN', 'TargetDSN', False, even_not_empty=True)

    # ### get DDLs
    @patch('pyvertica.migrate.subprocess.Popen', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._source_con', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    def test__get_ddls_empty(self, source, source_con, Popen):
        """
        Test :py:meth:`.VerticaMigrator._get_ddls`.
        None from DB, get from subprocess.
        """
        sql = 'CREATE TABLE cheese (id INT)'
        Popen.return_value.communicate.return_value = (sql, None)
        source.execute.return_value.fetchone.return_value = None
        ret = self.get_migrator()._get_ddls()
        self.assertEqual(ret, sql)

    @patch('pyvertica.migrate.subprocess.Popen', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._source_con', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    def test__get_ddls_empty_with_pwd(self, source, source_con, Popen):
        """
        Test :py:meth:`.VerticaMigrator._get_ddls`.
        None from DB, get from subprocess, password sent to vsql
        """
        sql = 'CREATE TABLE cheese (id INT)'
        Popen.return_value.communicate.return_value = (sql, None)
        source.execute.return_value.fetchone.return_value = None
        ret = self.get_migrator(source_pwd='tartiflette')._get_ddls()
        self.assertEqual(ret, sql)

    @patch('pyvertica.migrate.subprocess.Popen', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._source_con', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    def test__get_ddls_empty_with_exception(self, source, source_con, Popen):
        """
        Test :py:meth:`.VerticaMigrator._get_ddls`.
        None from DB, exception in subprocess.
        """
        source.execute.return_value.fetchone.return_value = None
        Popen.side_effect = CalledProcessError(42, 'Boom')
        self.assertRaises(VerticaMigratorError, self.get_migrator()._get_ddls)

    @patch('pyvertica.migrate.subprocess.Popen', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._source_con', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    def test__get_ddls_not_empty(self, source, source_con, Popen):
        """
        Test :py:meth:`.VerticaMigrator._get_ddls`.
        Result from DB.
        """
        Popen.return_value.communicate.return_value = ('something', None)
        ret = self.get_migrator()._get_ddls()
        self.assertEqual(ret, 'something')

    # ### SEQUENCE manipulation
    def test_sequence_regexp_valid(self):
        """
        Test regular expression VerticaMigrator._find_seq
        Test a valid syntax.
        """
        re = self.get_migrator()._find_seq.search
        seq = 'CREATE SEQUENCE schema.seq_name'
        m_seqs = re(seq)
        schema = m_seqs.group('schema')
        seqname = m_seqs.group('seq')
        self.assertEqual(schema, 'schema')
        self.assertEqual(seqname, 'seq_name')

    def test_sequence_regexp_invalid(self):
        """
        Test regular expression VerticaMigrator._find_seq
        Test an invalid syntax.
        """
        re = self.get_migrator()._find_seq.search
        seq = 'CREATE TABLE plop (id int)'
        m_seqs = re(seq)
        self.assertEqual(m_seqs, None)

    def test__is_sequence_true(self):
        """
        Test :py:meth:`.VerticaMigrator._is_sequence`,
        with a valid sequence
        """
        seq = self.get_migrator()._is_sequence(
            "CREATE SEQUENCE schema.seq_name"
        )
        self.assertEqual(seq, True)

    def test__is_sequence_false(self):
        """
        Test :py:meth:`.VerticaMigrator._is_sequence`,
        with a non-sequence DDL
        """
        seq = self.get_migrator()._is_sequence(
            "CREATE TABLE s.t (id INT)"
        )
        self.assertEqual(seq, False)

    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    def test_update_sequence_start(self, source):
        """
        Test :py:meth:`.VerticaMigrator._update_sequence_start`,
        """
        source.execute.return_value.fetchone.return_value = [42]
        seq = 'CREATE SEQUENCE schema.seq_name'
        res = self.get_migrator()._update_sequence_start(seq)
        self.assertEqual(res, seq + ' START WITH 43')

    # ### test IDENTITY

    def test_identity_regexp_valid(self):
        """
        Test regular expression VerticaMigrator._find_identity
        Test a valid syntax.
        """
        re = self.get_migrator()._find_identity.search
        ident = '''CREATE TABLE schema.cheese (
            id IDENTITY,
            runny INT)'''
        m_ids = re(ident)
        schema = m_ids.group('schema')
        table = m_ids.group('table')
        col = m_ids.group('col')
        self.assertEqual(schema, 'schema')
        self.assertEqual(table, 'cheese')
        self.assertEqual(col, 'id')

    def test_identity_regexp_invalid(self):
        """
        Test regular expression VerticaMigrator._find_identity
        Test an invalid syntax.
        """
        re = self.get_migrator()._find_identity.search
        ident = '''CREATE TABLE schema.cheese (
            id INT,
            runny INT)'''
        m_ids = re(ident)
        self.assertEqual(m_ids, None)

    def test__uses_identity_true(self):
        """
        """
        seq_true = self.get_migrator()._uses_identity(
            '''CREATE TABLE schema.cheese (
                id IDENTITY,
                runny BOOLEAN
                )
                '''
        )
        self.assertEqual(seq_true, True)

    def test__uses_identity_false(self):
        seq_false = self.get_migrator()._uses_identity(
            '''CREATE TABLE schema.cheese (
                id INT,
                runny BOOLEAN
                )
                '''
        )
        self.assertEqual(seq_false, False)

    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    def test__replace_initialised_identity(self, source):

        source.execute.return_value.fetchone.side_effect = [
        ['cheezy_seq'], [42]]

        ddl, new_id = self.get_migrator()._replace_identity(
            '''CREATE TABLE schema.cheese (
                id IDENTITY,
                runny BOOLEAN
                )
                '''
        )
        self.assertEqual(ddl, '''CREATE TABLE schema.cheese (
                id INT NOT NULL,
                runny BOOLEAN
                )
                ''')
        self.assertEqual(new_id, {'schema': 'schema',
                    'table': 'cheese',
                    'col': 'id',
                    'start':  43,
                    'name': 'cheezy_seq'
                    })

    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    def test__replace_uninitialised_identity(self, source):
        source.execute.return_value.fetchone.side_effect = [
        ['cheezy_seq'], [None]]
        ddl, new_id = self.get_migrator()._replace_identity(
            '''CREATE TABLE schema.cheese (
                id IDENTITY,
                runny BOOLEAN
                )
                '''
        )
        self.assertEqual(ddl, '''CREATE TABLE schema.cheese (
                id INT NOT NULL,
                runny BOOLEAN
                )
                ''')
        self.assertEqual(new_id, {'schema': 'schema',
                    'table': 'cheese',
                    'col': 'id',
                    'start':  1,
                    'name': 'cheezy_seq'
                    })

    # ### test TEMPORARY TABLE
    def test__is_tmptable_true(self):
        tmp_true = self.get_migrator()._is_temporary_table('''
            CREATE TEMPORARY TABLE schema.something (a INT)
            ''')
        self.assertEqual(tmp_true, True)

    def test__is_tmptable_false(self):
        tmp_false = self.get_migrator()._is_temporary_table('''
            CREATE TABLE schema.something
            blah
            ''')
        self.assertEqual(tmp_false, False)

    # ### test PROJECTION

    def test__is_proj_true(self):
        proj_true = self.get_migrator()._is_proj('''
            CREATE PROJECTION schema.something
            ( col1, col2) AS SELECT * FROM schema.table
            SEGMENTED BY hash(cols) ALL NODES OFFSET 0;
            ''')
        self.assertEqual(proj_true, True)

    def test__is_proj_false(self):
        proj_false = self.get_migrator()._is_proj('''
            CREATE TABLE schema.something
            blah
            ''')
        self.assertEqual(proj_false, False)

    # ### Get table list

    # get_table_list can not be unit tested in a useful way.

    ### Connection type
    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target', create=True)
    def test__connection_type(self, target, source):

        migrator = self.get_migrator()
        self.assertEqual('direct', migrator._connection_type())

        source.execute.side_effect = Exception('Cannot Connect')
        self.assertEqual('odbc', migrator._connection_type())

    ### DDL Migration

    def test__exec_ddl_already_exists(self):
        migrator = self.get_migrator(clever_ddls=True)
        migrator._commit = True
        migrator._target = Mock()
        migrator._target.execute.side_effect = pyodbc.ProgrammingError(
            '42601', 'message')
        migrator._exec_ddl('DDL')

    def test__exec_ddl_other_error(self):
        migrator = self.get_migrator(clever_ddls=True)
        migrator._commit = True
        migrator._target = Mock()
        migrator._target.execute.side_effect = pyodbc.ProgrammingError(
            '42', 'message')

        self.assertRaises(pyodbc.ProgrammingError, migrator._exec_ddl, 'DDL')

    def test_migrate_ddls_None(self):
        migrator = self.get_migrator()
        migrator._get_ddls = Mock(return_value=None)
        migrator._exec_ddl = Mock()
        migrator.migrate_ddls()
        self.assertEqual(0, migrator._exec_ddl.call_count)

    def test_migrate_ddls_empty(self):
        migrator = self.get_migrator()
        migrator._get_ddls = Mock(return_value='')
        migrator._exec_ddl = Mock()
        migrator.migrate_ddls()
        self.assertEqual(0, migrator._exec_ddl.call_count)

    def test_migrate_ddls_proj(self):
        migrator = self.get_migrator()
        migrator._get_ddls = Mock(return_value='CREATE PROJECTION')
        migrator._exec_ddl = Mock()
        migrator.migrate_ddls()
        self.assertEqual(0, migrator._exec_ddl.call_count)

    def test_migrate_ddls_tmptable(self):
        migrator = self.get_migrator()
        migrator._get_ddls = Mock(return_value='CREATE TEMPORARY TABLE')
        migrator._exec_ddl = Mock()
        migrator.migrate_ddls()
        self.assertEqual(0, migrator._exec_ddl.call_count)

    def test_migrate_ddls_sequence(self):
        migrator = self.get_migrator()
        migrator._get_ddls = Mock(return_value='not empty')
        migrator._is_proj = Mock(return_value=False)
        migrator._is_sequence = Mock(return_value=True)
        migrator._update_sequence_start = Mock(return_value='CREATE SEQ')
        migrator._uses_identity = Mock(return_value=False)
        migrator._exec_ddl = Mock()
        migrator.migrate_ddls()
        migrator._exec_ddl.assert_called_once_with(
            migrator._update_sequence_start.return_value)

    def test_migrate_ddls_identity(self):
        migrator = self.get_migrator()
        migrator._commit = True
        migrator._get_ddls = Mock(return_value='not empty')
        migrator._is_proj = Mock(return_value=False)
        migrator._is_sequence = Mock(return_value=False)
        migrator._uses_identity = Mock(return_value=True)
        migrator._replace_identity = Mock(return_value=['WITH IDENTITY',
         {'schema':'s', 'name':'n', 'start':'s', 'table':'t', 'col':'c'}])

        migrator._exec_ddl = Mock()
        migrator.migrate_ddls()
        self.assertEqual(call('WITH IDENTITY'),
            migrator._exec_ddl.call_args_list[0])
        self.assertEqual(3, migrator._exec_ddl.call_count)

    def test_migrate_ddls_error(self):
        migrator = self.get_migrator()
        migrator._commit = True
        migrator._get_ddls = Mock(return_value='CREATE VIEW')
        migrator._exec_ddl = Mock()
        migrator._exec_ddl.side_effect = Exception('Boom')

        self.assertRaises(VerticaMigratorError, migrator.migrate_ddls)
        # 1st call => error, go to errors
        # 2nd call => error again, errors do no shrink
        # 3rd call => to display
        self.assertEqual(3, migrator._exec_ddl.call_count)

    # ### Data migration
    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target', create=True)
    def test__migrate_table_direct(self, target, source):
        migrator = self.get_migrator()
        migrator._commit = True
        migrator._migrate_table('direct', 'a.table', {'db': 'db'})
        assert(migrator._source.execute.call_args_list[0][0][0].startswith(
            'EXPORT'))

    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target', create=True)
    @patch('pyvertica.migrate.VerticaBatch', Mock())
    def test__migrate_table_odbc(self, target, source):
        migrator = self.get_migrator()
        migrator._commit = True
        source.fetchone.side_effect = ['1', None]
        migrator._migrate_table('odbc', 'a.table', {'db': 'db'})
        assert(migrator._source.execute.call_args_list[0][0][0].startswith(
            'AT EPOCH'))

    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target', create=True)
    @patch('pyvertica.migrate.VerticaBatch', Mock())
    def test__migrate_table_other(self, target, source):
        migrator = self.get_migrator()
        self.assertRaises(VerticaMigratorError,
            migrator._migrate_table, 'something', 'a.table', {'db': 'db'})

    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target', create=True)
    def test_migrate_data_ok(self, target, source):
        migrator = self.get_migrator()
        migrator._connection_type = Mock()
        migrator._connection_details = Mock()
        migrator._get_table_list = Mock(return_value=['s.t'])
        migrator._migrate_table = Mock()
        self.assertEqual(migrator.migrate_data(''), None)

    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target', create=True)
    @patch('pyvertica.migrate.time.sleep', Mock())
    def test_migrate_data_error(self, target, source):
        migrator = self.get_migrator()
        migrator._connection_type = Mock()
        migrator._connection_details = Mock()
        migrator._get_table_list = Mock(return_value=['s.t'])
        migrator._migrate_table = Mock()
        migrator._migrate_table.side_effect = pyodbc.ProgrammingError(42, 'm')
        migrator.migrate_data('')
        self.assertEqual(migrator.migrate_data(''), None)

    @patch('pyvertica.migrate.VerticaMigrator._source', create=True)
    @patch('pyvertica.migrate.VerticaMigrator._target', create=True)
    @patch('pyvertica.migrate.time.sleep', Mock())
    def test_migrate_data_bigerror(self, target, source):
        migrator = self.get_migrator()
        migrator._connection_type = Mock()
        migrator._connection_details = Mock()
        migrator._get_table_list = Mock(return_value=['s.t', 'not.done'])
        migrator._migrate_table = Mock()
        migrator._migrate_table.side_effect = Exception(42, 'm')
        self.assertRaises(Exception, migrator.migrate_data, '')



from setuptools import setup

import pyvertica


setup(
    name='pyvertica',
    version=pyvertica.__version__,
    url='https://github.com/spilgames/pyvertica',
    license='BSD',
    author='Guillaume Roger, Orne Brocaar',
    author_email='datawarehouse@spilgames.com',
    description='Tools for performing batch imports into Vertica',
    long_description=open('README.rst').read(),
    packages=[
        'pyvertica',
        'pyvertica.tests',
        'pyvertica.tests.unit',
    ],
    scripts=[
        'scripts/vertica_batch_import',
        'scripts/vertica_migrate'
    ],
    install_requires=[
        'argparse',
        'logutils',
        'pyodbc==3.0.3',
    ]
)



import unittest2 as unittest

from mock import Mock, call, patch

from pyvertica.importer import (
    AlreadyImportedError,
    BaseImporter,
    BatchImportError,
)


class BaseImporterTestCase(unittest.TestCase):
    """
    Tests for :py:class:`.BaseImporter`.
    """
    def setUp(self):
        self.mapping_list = (
            {
                'field_name': 'field_1',
                'db_data_type': 'VARCHAR(10)',
            },
            {
                'field_name': 'field_2',
                'db_data_type': 'VARCHAR(20)',
                'db_field_name': 'db_field_2',
            },
            {
                'field_name': 'field_3',
                'db_data_type': 'VARCHAR(30)',
            }
        )

    def get_importer(self, **kwargs):
        arguments = {
            'reader_obj': Mock(),
            'odbc_kwargs': {'dsn': 'TestDSN'},
            'schema_name': 'schema',
            'batch_source_path': 'test/path',
        }
        arguments.update(kwargs)
        return BaseImporter(**arguments)

    def test___init__(self):
        """
        Test initialization of :py:class:`.BaseImporter`.
        """
        reader_obj = Mock()

        importer = self.get_importer(reader_obj=reader_obj, foo='bar')

        self.assertEqual(reader_obj, importer._reader_obj)
        self.assertEqual({'dsn': 'TestDSN'}, importer._odbc_kwargs)
        self.assertEqual('schema', importer._schema_name)
        self.assertEqual({
            'foo': 'bar',
            'batch_source_path': 'test/path',
        }, importer._kwargs)

    @patch('pyvertica.importer.BaseImporter._get_db_column_list')
    @patch('pyvertica.importer.VerticaBatch')
    def test__get_vertica_batch(self, VerticaBatch, get_db_column_list):
        """
        Test :py:meth:`.BaseImporter._get_vertica_batch`.
        """
        importer = self.get_importer()
        importer.table_name = 'test_table'

        batch_obj = importer._get_vertica_batch()

        self.assertEqual(VerticaBatch.return_value, batch_obj)
        VerticaBatch.assert_called_once_with(
            odbc_kwargs={'dsn': 'TestDSN'},
            table_name='schema.test_table',
            column_list=get_db_column_list.return_value,
        )

    def test__get_db_column_list(self):
        """
        Test :py:meth:`.BaseImporter._get_db_column_list`.
        """
        importer = self.get_importer()
        importer.mapping_list = self.mapping_list
        importer.extra_fields = (
            {'field_name': 'extra_1'},
            {'field_name': 'extra_2'},
        )

        self.assertEqual(
            ['extra_1', 'extra_2', 'field_1', 'db_field_2', 'field_3'],
            importer._get_db_column_list()
        )

    def test__get_row_value_list(self):
        """
        Test :py:meth:`.BaseImporter._get_row_value_list`.
        """
        importer = self.get_importer()
        importer.mapping_list = self.mapping_list
        importer.extra_fields = (
            {'field_name': 'extra1'},
            {'field_name': 'extra2'},
        )
        importer.get_extra_extra1_data = Mock(return_value='extra_data1')
        importer.get_extra_extra2_data = Mock(return_value='extra_data2')

        row_value_list = importer._get_row_value_list({
            'field_1': 'data1',
            'field_2': 'data2',
            'field_3': 'data3',
            'field_4': 'data4',
        })

        self.assertEqual(
            ['extra_data1', 'extra_data2', 'data1', 'data2', 'data3'],
            row_value_list
        )

    def test__insert_into_history(self):
        """
        Test :py:meth:`.BaseImporter._insert_into_history`.
        """
        importer = self.get_importer()
        importer.batch_source_name = Mock()
        importer.batch_source_type_name = Mock()
        importer.get_extra_batch_source_path_data = Mock()
        importer.get_extra_batch_import_timestamp_data = Mock()

        db_cursor = Mock()

        importer._insert_into_history(db_cursor)

        db_cursor.execute.assert_called_once_with(
            'INSERT INTO meta.batch_history (batch_source_name, '
            'batch_source_type_name, batch_source_path, '
            'batch_import_timestamp) VALUES (?, ?, ?, ?)',
            importer.batch_source_name,
            importer.batch_source_type_name,
            importer.get_extra_batch_source_path_data.return_value,
            importer.get_extra_batch_import_timestamp_data.return_value,
        )

    def test_get_sql_create_table_statement(self):
        """
        Test :py:meth:`.BaseImporter.get_sql_create_table_statement`.
        """
        importer = self.get_importer()
        importer.mapping_list = self.mapping_list
        importer.extra_fields = (
            {
                'field_name': 'extra_1',
                'db_data_type': 'INTEGER',
            },
            {
                'field_name': 'extra_2',
                'db_data_type': 'VARCHAR(25)',
            }
        )
        importer.table_name = 'test_table'

        self.assertEqual(
            'CREATE TABLE schema.test_table (extra_1 INTEGER, extra_2 '
            'VARCHAR(25), field_1 VARCHAR(10), db_field_2 VARCHAR(20), '
            'field_3 VARCHAR(30))',
            importer.get_sql_create_table_statement()
        )

    def test_start_import(self):
        """
        Test :py:meth:`.BaseImporter.start_import`.
        """
        batch_obj = Mock()
        batch_obj.get_errors.return_value = (False, Mock())
        batch_obj.get_cursor = Mock()

        importer = self.get_importer(reader_obj=[1, 2, 3])
        importer.get_batch_source_path_exists = Mock(return_value=False)
        importer._get_row_value_list = Mock(side_effect=['a', 'b', 'c'])
        importer._get_vertica_batch = Mock(return_value=batch_obj)
        importer._insert_into_history = Mock()

        importer.start_import()

        importer._get_vertica_batch.assert_called_once_with()
        importer.get_batch_source_path_exists.assert_called_once_with(
            'test/path', odbc_kwargs={'dsn': 'TestDSN'})
        self.assertEqual(
            [call(1), call(2), call(3)],
            importer._get_row_value_list.call_args_list
        )
        self.assertEqual(
            [call('a'), call('b'), call('c')],
            batch_obj.insert_list.call_args_list
        )
        batch_obj.get_errors.assert_called_once_with()
        batch_obj.get_cursor.assert_called_once_with()
        importer._insert_into_history.assert_called_once_with(
            batch_obj.get_cursor.return_value)
        batch_obj.commit.assert_called_once_with()

    @patch('pyvertica.importer.logger')
    def test_start_import_errors(self, logger):
        """
        Test :py:meth:`.BaseImporter.start_import` with errors.
        """
        batch_obj = Mock()
        batch_obj.get_errors.return_value = (True, ['Error 1\n', 'Error 2\r'])

        importer = self.get_importer(reader_obj=[])
        importer._kwargs['batch_source_path'] = 'test/path'
        importer.get_batch_source_path_exists = Mock(return_value=False)
        importer.mapping_list = self.mapping_list
        importer._get_vertica_batch = Mock(return_value=batch_obj)

        self.assertRaises(BatchImportError, importer.start_import)

        self.assertEqual(0, batch_obj.commit.call_count)

        self.assertEqual([
            call('Batch error (test/path): Error 1'),
            call('Batch error (test/path): Error 2'),
        ], logger.error.call_args_list)

    def test_start_import_already_imported(self):
        """
        Test :py:meth:`.BaseImporter.start_import` with already imported record
        """
        importer = self.get_importer()
        importer.get_batch_source_path_exists = Mock(return_value=True)
        importer._kwargs['batch_source_path'] = 'test/path'

        self.assertRaises(AlreadyImportedError, importer.start_import)

    @patch('pyvertica.importer.BaseImporter.batch_source_name', 'test_bsn')
    @patch('pyvertica.importer.BaseImporter.batch_source_type_name', 'ga3')
    @patch('pyvertica.importer.get_connection')
    def test_get_batch_source_path_exists_true(self, get_connection):
        """
        Test :py:meth:`.BaseImporter.get_batch_source_path_exists`.

        This tests the case of ``True``.
        """
        connection = get_connection.return_value
        cursor = connection.cursor.return_value
        cursor.fetchone.return_value = ['test/path']

        self.assertTrue(
            BaseImporter.get_batch_source_path_exists(
                'test/path', odbc_kwargs={'dsn': 'TestDSN'})
        )

        get_connection.assert_called_once_with(dsn='TestDSN')
        connection.cursor.assert_called_once_with()
        cursor.execute.assert_called_once_with(
            'SELECT batch_source_path FROM meta.batch_history '
            'WHERE batch_source_name = ? AND batch_source_type_name = ? AND '
            'batch_source_path = ? LIMIT 1',
            'test_bsn',
            'ga3',
            'test/path'
        )
        cursor.fetchone.assert_called_once_with()

    @patch('pyvertica.importer.get_connection')
    def test_get_batch_source_path_exists_false(self, get_connection):
        """
        Test :py:meth:`.BaseImporter.get_batch_source_path_exists`.

        This tests the case of ``False``.
        """
        connection = get_connection.return_value
        cursor = connection.cursor.return_value
        cursor.fetchone.return_value = None

        self.assertFalse(
            BaseImporter.get_batch_source_path_exists(
                'test/path', odbc_kwargs={'dsn': 'TestDSN'})
        )

    @patch('pyvertica.importer.BaseImporter.batch_source_name', 'test_bsn')
    @patch('pyvertica.importer.BaseImporter.batch_source_type_name', 'ga3')
    @patch('pyvertica.importer.get_connection')
    def test_get_last_imported_batch_source_path(self, get_connection):
        """
        Test :py:meth:`.BaseImporter.get_last_imported_batch_souce_path`.
        """
        connection = get_connection.return_value
        cursor = connection.cursor.return_value
        cursor.fetchone.return_value = ('foo/bar',)

        self.assertEqual(
            'foo/bar',
            BaseImporter.get_last_imported_batch_source_path(
                odbc_kwargs={'dsn': 'TestDSN'})
        )

        get_connection.assert_called_once_with(dsn='TestDSN')
        connection.cursor.assert_called_once_with()
        cursor.execute.assert_called_once_with(
            'SELECT batch_source_path FROM meta.batch_history '
            'WHERE batch_source_name = ? AND batch_source_type_name = ? '
            'ORDER BY batch_import_timestamp DESC LIMIT 1',
            'test_bsn',
            'ga3'
        )

    @patch('pyvertica.importer.get_connection')
    def test_get_last_imported_batch_source_path_none(self, get_connection):
        """
        Test :py:meth:`.BaseImporter.get_last_imported_batch_source_path`.

        This tests the case of returning ``None``.
        """
        connection = get_connection.return_value
        cursor = connection.cursor.return_value
        cursor.fetchone.return_value = None

        self.assertEqual(
            None, BaseImporter.get_last_imported_batch_source_path(
                odbc_kwargs={'dsn': 'TestDSN'}))

    def test_get_extra_batch_source_name_data(self):
        """
        Test :py:meth:`.BaseImporter.get_extra_batch_source_name_data`.
        """
        importer = self.get_importer()
        importer.batch_source_name = 'foo_bar'

        self.assertEqual(
            'foo_bar',
            importer.get_extra_batch_source_name_data({})
        )

    def test_get_extra_batch_source_path_data(self):
        """
        Test :py:meth:`.BaseImporter.get_extra_batch_source_path_data`.
        """
        importer = self.get_importer()
        importer._kwargs = {'batch_source_path': 'batch/source/path'}

        self.assertEqual(
            'batch/source/path', importer.get_extra_batch_source_path_data({}))

    @patch('pyvertica.importer.datetime')
    def test_get_extra_batch_import_timestamp_data(self, datetime):
        """
        Test :py:meth:`.BaseImporter.get_extra_batch_import_timestamp_data`.
        """
        utc_datetime = Mock()
        datetime.utcnow.return_value = utc_datetime

        importer = self.get_importer()

        self.assertEqual(None, importer._batch_import_timestamp)
        self.assertEqual(
            utc_datetime.isoformat.return_value,
            importer.get_extra_batch_import_timestamp_data({})
        )
        utc_datetime.isoformat.assert_called_once_with(' ')
        self.assertEqual(
            utc_datetime.isoformat.return_value,
            importer._batch_import_timestamp
        )
